#!/usr/bin/perl
#
# $Id$
#
# RDGraph Graphing and Syslog Uptake Processor
#  mprigge@gmail.com
#  pvenezia@jpj.net
#

use strict;
use RRDs;
use Data::Dumper;
use Hash::Merge qw( merge );
use Time::HiRes qw(setitimer getitimer ITIMER_REAL);
use Getopt::Long;
use Unix::Syslog qw(:subs :macros);

my $C = {
  'cmdline' => $0." ".join(' ',@ARGV), 
  'debug' => 0,
  'mode' => 'syslog',
  'fifo' => '/var/run/rdgraph.fifo',
  'pidfile' => '/var/run/rdgraph.pid',
  'base' => {
    'rrdDir' => '/var/rrd/rdgraph',
    'imgDir' => '/var/www/html/relaydelay/rrdimages',
  },
  'monitors' => {
    'YOUR-COMPANY' => {
      'enabled' => 1,
      'name' => 'YOUR-COMPANY',
      'hosts' => [ 'RELAYDELAYHOSTNAME', ],
      'domains' => [ 'YOUR-COMPANY.com' ],
    },
    '__default' => {
      'enabled' => 0,
      'name' => 'Default_RRD',
      'graphs' => {
        'BlockedVsPassed' => {
           'name' => "BlockedVsPassed",
           'periods' => [ 'now-4hours','now-24hours','now-7days','now-4weeks','now-12months','now-2years', ],
           'desc' => "Graph of Passed Mail vs. Blocked Mail", 
           'callback' => \&__graphBlockedVsPassed,
        },
        'BlockedDetail' => {
           'name' => "BlockedDetail",
           'periods' => [ 'now-4hours','now-24hours','now-7days','now-4weeks','now-12months','now-2years', ],
           'desc' => "Graph detailing Blocked Mail",
           'callback' => \&__graphBlockedDetail,
        },
        'PassedDetail' => {
           'name' => "PassedDetail",
           'periods' => [ 'now-4hours','now-24hours','now-7days','now-4weeks','now-12months','now-2years', ],
           'desc' => "Graph detailing Passed Mail",
           'callback' => \&__graphPassedDetail,
        },

      },
      'hosts' => [ '.*', ],
      'domains' => [ '.*', ],
      'step' => 60,
      'counter' => {},
      'DS' => {
        "BLOCK" => {
           "type" => "COUNTER",
           "name" => "BLOCK",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "BLOCK_NEW_MSG" => {
           "type" => "COUNTER",
           "name" => "BLOCK_NEW_MSG",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "BLOCK_BL_RELAY" => {
           "type" => "COUNTER",
           "name" => "BLOCK_BL_RELAY",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "BLOCK_BL_DOMAIN" => {
           "type" => "COUNTER",
           "name" => "BLOCK_BL_DOMAIN",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "BLOCK_BL_RCPT" => {
           "type" => "COUNTER",
           "name" => "BLOCK_BL_RCPT",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "BLOCK_BK_NOT_EXP" => {
           "type" => "COUNTER",
           "name" => "BLOCK_BK_NOT_EXP",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "PASS" => {
           "type" => "COUNTER",
           "name" => "PASS",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "PASS_WL_DOMAIN" => {
           "type" => "COUNTER",
           "name" => "PASS_WL_DOMAIN",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "PASS_WL_RELAY" => {
           "type" => "COUNTER",
           "name" => "PASS_WL_RELAY",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "PASS_WL_RCPT" => {
           "type" => "COUNTER",
           "name" => "PASS_WL_RCPT",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
        "PASS_BK_EXP" => {
           "type" => "COUNTER",
           "name" => "PASS_BK_EXP",
           "heartbeat" => 120,
           "min" => 0,
           "max" => undef,
        },
      },
      'RRA' => {
        "1Day" => {
           "type" => "AVERAGE",
           "xff" => "0.5",
           "steps" => "1",
           "rows" => "1440",
        },
        "1Week" => {
           "type" => "AVERAGE",
           "xff" => "0.40",
           "steps" => "5",
           "rows" => "2016",
        },
        "4Week" => {
           "type" => "AVERAGE",
           "xff" => "0.1",
           "steps" => "30",
           "rows" => "1344",
        },
        "52Week" => {
           "type" => "AVERAGE",
           "xff" => "0.1",
           "steps" => "120",
           "rows" => "4368",
        },
      },
    },
  },
};

GetOptions ("mode=s"   => \$C->{"mode"},
            "debug"  => \$C->{"debug"});

sub Parse; # Predeclare &Parse so that the fork subroutine doesnt barf

&Main;

sub Main {
  &buildConfig;
  &initRRD;

  if($C->{"mode"} eq "syslog") {
    $0 = "rdgraph.pl: Daemonizing...";

    # Detach from console and start parsing (and use a hammer)
    unless(fork()) {
      unless(fork()) {
        #LURK: while(1) {
        #  sleep(1);
        #  last LURK if getppid() == 1;
        #}
        &ParseInit;
      }
      exit;
    }
    wait();
    exit;

  } elsif($C->{"mode"} eq "graph") {
    $0 = "rdgraph.pl: Graphing...";
    &Graph;

  } else {
    &Usage;
  }
}

sub ParseInit {
  
  openlog("rdgraph",LOG_PID,LOG_LOCAL3);
  syslog(LOG_NOTICE,"Starting rdgraph log processing... (".$C->{cmdline}.")");

  &BreakIfRunning;

  $SIG{'HUP'} = \&Restart;
  $SIG{'INT'} = \&Shutdown;
  $SIG{'TERM'} = \&Shutdown;
  $SIG{'ALRM'} = \&Write;

  setitimer(ITIMER_REAL,$C->{monitors}->{__default}->{step},$C->{monitors}->{__default}->{step});

  &Parse;
}

# Usage
sub Usage {
  #TODO
  print "OOPS!!\n";
}


# Graph Writer
sub Graph {
  print STDERR "Graph() called...\n" if($C->{"debug"});

  foreach my $monitor (values %{$C->{"monitors"}}) {
    next if($monitor->{enabled} != 1);

    print STDERR $monitor->{name}."\n" if($C->{"debug"});

    foreach my $graph (values %{$monitor->{graphs}}) {
      $graph->{callback}->($monitor,$graph);
    }
  }
}

sub __graphBlockedVsPassed($) {
  my $monitor = shift(@_);
  my $graph = shift(@_);

  print "__graphBlockedVsPassed() called for ".$monitor->{name}."/".$graph->{name}."\n" if($C->{debug});

  if(!-d $C->{base}->{imgDir}."/".$monitor->{'name'}) {
    mkdir($C->{base}->{imgDir}."/".$monitor->{'name'});
  }


  foreach my $period (@{$graph->{periods}}) {

    my $imgfile = $C->{base}->{imgDir}."/".$monitor->{'name'}."/".$graph->{name}."_".$period.".png";

         RRDs::graph(
           $imgfile,
           '--start='.$period,
           '--end=now',
           "--title=".$graph->{desc}." for ".$monitor->{name},
           '--vertical-label=Messages per Minute',
           '--width=600',
           '--height=150',
           'DEF:BLOCK='.$monitor->{'filename'}.':BLOCK:AVERAGE',
           'DEF:PASS='.$monitor->{'filename'}.':PASS:AVERAGE',

           'CDEF:BLOCK_M=BLOCK,60,*',
           'CDEF:PASS_M=PASS,60,*',
           'CDEF:BLOCK_HR=BLOCK,3600,*',
           'CDEF:PASS_HR=PASS,3600,*',

           'CDEF:BLOCK_STK=BLOCK_M,PASS_M,+',
           'CDEF:WIPEOUT=BLOCK,PASS,+,UN,INF,UNKN,IF',

           'COMMENT:  ',
           'AREA:BLOCK_STK#FF0000:Blocked Messages',
           'COMMENT:       ',
           'GPRINT:BLOCK_M:MAX:Max\: %lf',
           'GPRINT:BLOCK_M:LAST:Cur\: %lf',
           'GPRINT:BLOCK_M:AVERAGE:Avg\: %lf',
           'GPRINT:BLOCK_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT:  ',
           'AREA:PASS_M#00FF00:Passed Messages',
           'COMMENT:        ',
           'GPRINT:PASS_M:MAX:Max\: %lf',
           'GPRINT:PASS_M:LAST:Cur\: %lf',
           'GPRINT:PASS_M:AVERAGE:Avg\: %lf',
           'GPRINT:PASS_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',
           'COMMENT:\n',
           'COMMENT:  ',
           'AREA:WIPEOUT#D9D9D9:(no data)\n',
           'COMMENT:All figures in messages per minute unless otherwise specified.\c',
           'COMMENT:Generated on: '.localtime(time()).'\c'
         );
    my $ERR=RRDs::error;
    warn "ERROR while updating ".$monitor->{name}.": $ERR\n" if $ERR;

  }

}


sub __graphBlockedDetail($) {
  my $monitor = shift(@_);
  my $graph = shift(@_);

  print "__graphBlockedDetail() called for ".$monitor->{name}."/".$graph->{"name"}."\n" if($C->{debug});

  if(!-d $C->{base}->{imgDir}."/".$monitor->{'name'}) {
    mkdir($C->{base}->{imgDir}."/".$monitor->{'name'});
  }


  foreach my $period (@{$graph->{periods}}) {

    my $imgfile = $C->{base}->{imgDir}."/".$monitor->{'name'}."/".$graph->{"name"}."_".$period.".png";


         RRDs::graph(
           $imgfile,
           '--start='.$period,
           '--end=now',
           "--title=".$graph->{desc}." for ".$monitor->{name},
           '--vertical-label=Messages per Minute',
           '--width=600',
           '--height=150',

           'DEF:NEW_MSG='.$monitor->{'filename'}.':BLOCK_NEW_MSG:AVERAGE',
           'DEF:BK_NOT_EXP='.$monitor->{'filename'}.':BLOCK_BK_NOT_EXP:AVERAGE',
           'DEF:BL_RELAY='.$monitor->{'filename'}.':BLOCK_BL_RELAY:AVERAGE',
           'DEF:BL_DOMAIN='.$monitor->{'filename'}.':BLOCK_BL_DOMAIN:AVERAGE',
           'DEF:BL_RCPT='.$monitor->{'filename'}.':BLOCK_BL_RCPT:AVERAGE',

           'CDEF:BL_RCPT_M=BL_RCPT,60,*',
           'CDEF:BL_DOMAIN_M=BL_DOMAIN,60,*',
           'CDEF:BL_RELAY_M=BL_RELAY,60,*',
           'CDEF:BK_NOT_EXP_M=BK_NOT_EXP,60,*',
           'CDEF:NEW_MSG_M=NEW_MSG,60,*',

           'CDEF:BL_RCPT_STK=NEW_MSG_M,BK_NOT_EXP_M,BL_RELAY_M,BL_DOMAIN_M,BL_RCPT_M,+,+,+,+',
           'CDEF:BL_DOMAIN_STK=NEW_MSG_M,BK_NOT_EXP_M,BL_RELAY_M,BL_DOMAIN_M,+,+,+',
           'CDEF:BL_RELAY_STK=NEW_MSG_M,BK_NOT_EXP_M,BL_RELAY_M,+,+',
           'CDEF:BK_NOT_EXP_STK=NEW_MSG_M,BK_NOT_EXP_M,+',
           'CDEF:WIPEOUT=NEW_MSG_M,BK_NOT_EXP_M,BL_RELAY_M,BL_DOMAIN_M,BL_RCPT_M,+,+,+,+,UN,INF,UNKN,IF',

           'CDEF:BL_RCPT_HR=BL_RCPT,3600,*',
           'CDEF:BL_DOMAIN_HR=BL_DOMAIN,3600,*',
           'CDEF:BL_RELAY_HR=BL_RELAY,3600,*',
           'CDEF:BK_NOT_EXP_HR=BK_NOT_EXP,3600,*',
           'CDEF:NEW_MSG_HR=NEW_MSG,3600,*',

           'COMMENT: ',
           'AREA:BL_RCPT_STK#00FF88:Blacklisted by Recipient ',
           'GPRINT:BL_RCPT_M:MAX:Max\: %lf',
           'GPRINT:BL_RCPT_M:LAST:Cur\: %lf',
           'GPRINT:BL_RCPT_M:AVERAGE:Avg\: %lf',
           'GPRINT:BL_RCPT_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT: ',
           'AREA:BL_DOMAIN_STK#0088FF:Blacklisted by Domain',
           'COMMENT:  ',
           'GPRINT:BL_DOMAIN_M:MAX:Max\: %lf',
           'GPRINT:BL_DOMAIN_M:LAST:Cur\: %lf',
           'GPRINT:BL_DOMAIN_M:AVERAGE:Avg\: %lf',
           'GPRINT:BL_DOMAIN_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT: ',
           'AREA:BL_RELAY_STK#00FFFF:Blacklisted by Relay',
           'COMMENT:   ',
           'GPRINT:BL_RELAY_M:MAX:Max\: %lf',
           'GPRINT:BL_RELAY_M:LAST:Cur\: %lf',
           'GPRINT:BL_RELAY_M:AVERAGE:Avg\: %lf',
           'GPRINT:BL_RELAY_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT: ',
           'AREA:BK_NOT_EXP_STK#FF8800:Block Not Expired',
           'COMMENT:      ',
           'GPRINT:BK_NOT_EXP_M:MAX:Max\: %lf',
           'GPRINT:BK_NOT_EXP_M:LAST:Cur\: %lf',
           'GPRINT:BK_NOT_EXP_M:AVERAGE:Avg\: %lf',
           'GPRINT:BK_NOT_EXP_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT: ',
           'AREA:NEW_MSG_M#FF0000:New Message',
           'COMMENT:            ',
           'GPRINT:NEW_MSG_M:MAX:Max\: %lf',
           'GPRINT:NEW_MSG_M:LAST:Cur\: %lf',
           'GPRINT:NEW_MSG_M:AVERAGE:Avg\: %lf',
           'GPRINT:NEW_MSG_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT:\n',
           'COMMENT: ',
           'AREA:WIPEOUT#D9D9D9:(no data)\n',
           'COMMENT:All figures in messages per minute unless otherwise specified.\c',
           'COMMENT:Generated on: '.localtime(time()).'\c'
         );
    my $ERR=RRDs::error;
    warn "ERROR while updating ".$monitor->{name}.": $ERR\n" if $ERR;

  }

}


sub __graphPassedDetail($) {
  my $monitor = shift(@_);
  my $graph = shift(@_);

  print "__graphPassedDetail() called for ".$monitor->{name}."/".$graph->{"name"}."\n" if($C->{debug});

  if(!-d $C->{base}->{imgDir}."/".$monitor->{'name'}) {
    mkdir($C->{base}->{imgDir}."/".$monitor->{'name'});
  }


  foreach my $period (@{$graph->{periods}}) {

    my $imgfile = $C->{base}->{imgDir}."/".$monitor->{'name'}."/".$graph->{"name"}."_".$period.".png";

         RRDs::graph(
           $imgfile,
           '--start='.$period,
           '--end=now',
           "--title=".$graph->{desc}." for ".$monitor->{name},
           '--vertical-label=Messages per Minute',
           '--width=600',
           '--height=150',

           'DEF:BK_EXP='.$monitor->{'filename'}.':PASS_BK_EXP:AVERAGE',
           'DEF:WL_RELAY='.$monitor->{'filename'}.':PASS_WL_RELAY:AVERAGE',
           'DEF:WL_DOMAIN='.$monitor->{'filename'}.':PASS_WL_DOMAIN:AVERAGE',
           'DEF:WL_RCPT='.$monitor->{'filename'}.':PASS_WL_RCPT:AVERAGE',

           'CDEF:WL_RCPT_M=WL_RCPT,60,*',
           'CDEF:WL_DOMAIN_M=WL_DOMAIN,60,*',
           'CDEF:WL_RELAY_M=WL_RELAY,60,*',
           'CDEF:BK_EXP_M=BK_EXP,60,*',
           'CDEF:WL_RCPT_HR=WL_RCPT,3600,*',
           'CDEF:WL_DOMAIN_HR=WL_DOMAIN,3600,*',
           'CDEF:WL_RELAY_HR=WL_RELAY,3600,*',
           'CDEF:BK_EXP_HR=BK_EXP,3600,*',


           'CDEF:WL_RCPT_STK=BK_EXP_M,WL_RELAY_M,WL_DOMAIN_M,WL_RCPT_M,+,+,+',
           'CDEF:WL_DOMAIN_STK=BK_EXP_M,WL_RELAY_M,WL_DOMAIN_M,+,+',
           'CDEF:WL_RELAY_STK=BK_EXP_M,WL_RELAY_M,+',
           'CDEF:WIPEOUT=BK_EXP_M,WL_RELAY_M,WL_DOMAIN_M,WL_RCPT_M,+,+,+,UN,INF,UNKN,IF',

           'COMMENT:  ',
           'AREA:WL_RCPT_STK#00FF88:Whitelisted by Recipient',
           'COMMENT: ',
           'GPRINT:WL_RCPT_M:MAX:Max\: %lf',
           'GPRINT:WL_RCPT_M:LAST:Cur\: %lf',
           'GPRINT:WL_RCPT_M:AVERAGE:Avg\: %lf',
           'GPRINT:WL_RCPT_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT:  ',
           'AREA:WL_DOMAIN_STK#0088FF:Whitelisted by Domain',
           'COMMENT:    ',
           'GPRINT:WL_DOMAIN_M:MAX:Max\: %lf',
           'GPRINT:WL_DOMAIN_M:LAST:Cur\: %lf',
           'GPRINT:WL_DOMAIN_M:AVERAGE:Avg\: %lf',
           'GPRINT:WL_DOMAIN_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT:  ',
           'AREA:WL_RELAY_STK#00FFFF:Whitelisted by Relay',
           'COMMENT:     ',
           'GPRINT:WL_RELAY_M:MAX:Max\: %lf',
           'GPRINT:WL_RELAY_M:LAST:Cur\: %lf',
           'GPRINT:WL_RELAY_M:AVERAGE:Avg\: %lf',
           'GPRINT:WL_RELAY_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT:  ',
           'AREA:BK_EXP_M#FF8800:Block Expired',
           'COMMENT:            ',
           'GPRINT:BK_EXP_M:MAX:Max\: %lf',
           'GPRINT:BK_EXP_M:LAST:Cur\: %lf',
           'GPRINT:BK_EXP_M:AVERAGE:Avg\: %lf',
           'GPRINT:BK_EXP_HR:AVERAGE:msg/hr\: %lf',
           'COMMENT:\n',

           'COMMENT:\n',
           'COMMENT:  ',
           'AREA:WIPEOUT#D9D9D9:(no data)\n',
           'COMMENT:All figures in messages per minute unless otherwise specified.\c',
           'COMMENT:Generated on: '.localtime(time()).'\c'
         );
    my $ERR=RRDs::error;
    warn "ERROR while updating ".$monitor->{name}.": $ERR\n" if $ERR;

  }

}


# RRD Writer
sub Write {
  $0 = "rdgraph.pl: Writing RRD data...";
  print STDERR "Write() called...\n" if($C->{"debug"});
  foreach my $monitor (values %{$C->{"monitors"}}) {
    next if($monitor->{enabled} != 1);

    print STDERR $monitor->{name}."\n" if($C->{"debug"});
    print STDERR Dumper($monitor->{counter}) if($C->{"debug"});

    my(@keys,@values);
    foreach my $key (keys %{$monitor->{counter}}) {
      push(@keys,$key);
      push(@values,$monitor->{counter}->{$key});
    }

    my $ERR;

    RRDs::update($monitor->{filename},"--template",join(":",@keys),join(":",time(),@values));
    $ERR=RRDs::error;
    warn "ERROR while updating ".$monitor->{name}.": $ERR\n" if $ERR;

    #$R->{graphingCallback}->($H,$R);
    #$ERR=RRDs::error;
    #die "ERROR while graphing ".$H->{hostname}."/".$test.": $ERR\n" if $ERR;

  }

  $0 = "rdgraph.pl: Parsing syslog stream..."; 
}


# Log Processor
sub Parse {
  $0 = "rdgraph.pl: Parsing syslog stream..."; 

  # Open log FIFO
  open(FIFO,"<".$C->{fifo}) or die("Could not open log FIFO: $!");

  # Commence log reading
  while(my $line = <FIFO>) {
    chomp($line);

    # Create entry bucket
    my $e = {};

    # Parse incoming line
    if($line =~ /^(\w+\s+\d+\s+\d+\:\d+\:\d+)\s+(\S+)\s+rdgraph\[\d+\]\:\s+([^\,]+)\,([^\,]+)\,([^\,]+)\,([^\,]+)\,([^\,]+)$/i) {
      $e = {
        "datetime" => $1,
        "host" => $2,
        "condition" => $3,
        "subcondition" => $4,
        "mail_from" => $5,
        "rcpt_to" => $6,
        "relay_ip" => $7,
      };

    } else {
      print STDERR "WARNING: Entry ignored ($line)\n" if($C->{"debug"});
      next;
    }

    foreach my $monitor (values %{$C->{monitors}}) {
      next unless($monitor->{enabled} == 1);

      my $hstrgx = join("\|",@{$monitor->{hosts}});
      if($e->{host} !~ /^($hstrgx)$/) {
        next;
      };


      my $domrgx = join("\|",@{$monitor->{domains}});
      if($e->{rcpt_to} !~ /($domrgx)\>$/) { 
        next;
      };


      if(defined($monitor->{counter}->{$e->{"condition"}})) {
        $monitor->{counter}->{$e->{"condition"}}++;
      }
      if(defined($monitor->{counter}->{$e->{"condition"}."_".$e->{"subcondition"}})) {
        $monitor->{counter}->{$e->{"condition"}."_".$e->{"subcondition"}}++;
      }

    }       
  }
}


sub buildConfig {

         Hash::Merge::specify_behavior( {
               SCALAR => {
                       SCALAR => sub { $_[1] },
                       ARRAY  => sub { [ $_[0], @{$_[1]} ] },
                       HASH   => sub { $_[1] } },
               ARRAY => {
                       SCALAR => sub { $_[1] },
                       ARRAY  => sub { [ @{$_[1]}, ], }, # modified to overwrite arrays
                       HASH   => sub { $_[1] } },
               HASH => {
                       SCALAR => sub { $_[1] },
                       ARRAY  => sub { [ values %{$_[0]}, @{$_[1]} ] },
                       HASH   => sub { Hash::Merge::_merge_hashes( $_[0], $_[1] ) } }
         }, "My Behavior" );


  foreach my $monitor (keys %{$C->{"monitors"}}) {
    next if($monitor eq '__default');

    my $cfg = merge($C->{"monitors"}->{"__default"},$C->{"monitors"}->{$monitor});
    $C->{"monitors"}->{$monitor} = $cfg;

    foreach my $DS (values %{$C->{"monitors"}->{$monitor}->{"DS"}}) {
      $C->{"monitors"}->{$monitor}->{"counter"}->{$DS->{name}} = 0;
    }

  }
}


sub initRRD() {
  my $start=time;

  foreach my $monitor (values %{$C->{monitors}}) {
    if(!$monitor->{"enabled"}) { next; }

    $monitor->{filename} = $C->{base}->{rrdDir}."/".$monitor->{name}.".rrd";

    if(! -f $monitor->{filename}) {
      my @DSs;
      foreach my $ds (values %{$monitor->{DS}}) {
        if(!defined($ds->{min})) {
          $ds->{min} = "U";
        }
        if(!defined($ds->{max})) {
          $ds->{max} = "U";
        }

        push(@DSs,join(":",("DS",$ds->{name},$ds->{type},$ds->{heartbeat},$ds->{min},$ds->{max})));
      }
      my @RRAs;
      foreach my $rra (values %{$monitor->{RRA}}) {
        push(@RRAs,join(":",("RRA",$rra->{type},$rra->{xff},$rra->{steps},$rra->{rows})));
      }

      RRDs::create($monitor->{filename}, "--start",$start-1, "--step",$monitor->{step},@DSs,@RRAs);

      my $ERROR = RRDs::error;
      die "$0: unable to create `".$monitor->{filename}."': $ERROR\n" if $ERROR;
    }
  }
}


sub BreakIfRunning() {
    my $writepid = 1;

    sleep(1);

    if (open (PIDFILE, $C->{pidfile})) {
        my ($pid) = <PIDFILE>;
        chomp $pid;

        if($pid =~ /\d+/) {
            if($pid == $$) {
                $writepid = 0;
            } elsif( kill(0, $pid) == 1 ) {
                syslog(LOG_ERR,"Another instance (PID: %d) already running.. Exiting.",$pid);
                print "rdgraph: Another instance is already running at pid $pid. Perhaps you meant to \`kill -HUP $pid\`.\n";
                exit(1);
            }
        }
    }

    if ($writepid) {
        open (PIDFILE, ">".$C->{pidfile}) or syslog(LOG_ERR,"Unable to open %s for writing.",$C->{pidfile});
        print PIDFILE "$$\n";
        print PIDFILE "# started at " . scalar(localtime()) . "\n";
        close PIDFILE;
    } else {
        open (PIDFILE, ">>".$C->{pidfile}) or syslog(LOG_ERR,"Unable to open %s for appending.",$C->{pidfile});
        print PIDFILE "# restarted at " . scalar(localtime()) . "\n";
        close PIDFILE;
    }
}

sub Restart($) {
    my $sig = shift(@_);
    syslog(LOG_NOTICE,"Restarting on SIG%s",$sig);
    exec($C->{cmdline});
}

sub Shutdown($) {
    my $sig = shift(@_);
    syslog(LOG_NOTICE,"Shutting down on SIG%s",$sig);
    unlink($C->{pidfile});
    exit(0);
}

